@@ -39,14 +39,32 @@
 
 #include <glib/gi18n-lib.h>
 
+#define DTLS_MESSAGE_MAX_SIZE 65536
+
 typedef struct _GTlsConnectionOpensslPrivate
 {
   BIO *bio;
+  guint8 *dtls_rx;
+  guint8 *dtls_tx;
   GMutex ssl_mutex;
 
   gboolean shutting_down;
 } GTlsConnectionOpensslPrivate;
 
+typedef int (*GTlsOpensslIOFunc) (SSL *ssl, gpointer user_data);
+
+typedef struct _ReadRequest
+{
+  void *buffer;
+  gsize count;
+} ReadRequest;
+
+typedef struct _WriteRequest
+{
+  const void *buffer;
+  gsize count;
+} WriteRequest;
+
 static void g_tls_connection_openssl_initable_iface_init (GInitableIface *iface);
 
 G_DEFINE_ABSTRACT_TYPE_WITH_CODE (GTlsConnectionOpenssl, g_tls_connection_openssl, G_TYPE_TLS_CONNECTION_BASE,
@@ -62,6 +80,8 @@ g_tls_connection_openssl_finalize (GObject *object)
 
   priv = g_tls_connection_openssl_get_instance_private (openssl);
 
+  g_free (priv->dtls_rx);
+  g_free (priv->dtls_tx);
   g_mutex_clear (&priv->ssl_mutex);
 
   G_OBJECT_CLASS (g_tls_connection_openssl_parent_class)->finalize (object);
@@ -218,9 +238,6 @@ end_openssl_io (GTlsConnectionOpenssl  *openssl,
 
   if (my_error)
     g_propagate_error (error, my_error);
-  else
-    /* FIXME: this is just for debug */
-    g_message ("end_openssl_io %s: %d, %d, %d", G_IS_TLS_CLIENT_CONNECTION (openssl) ? "client" : "server", err_code, err_lib, reason);
 
   if (ret == 0 && err == 0 && err_lib == 0 && err_code == SSL_ERROR_SYSCALL
       && (direction == G_IO_IN || direction == G_IO_OUT))
@@ -241,7 +258,113 @@ end_openssl_io (GTlsConnectionOpenssl  *openssl,
   return G_TLS_CONNECTION_BASE_ERROR;
 }
 
-#if OPENSSL_VERSION_NUMBER >= 0x10002000L || defined (LIBRESSL_VERSION_NUMBER)
+static GTlsConnectionBaseStatus
+perform_openssl_io (GTlsConnectionOpenssl  *openssl,
+                    GIOCondition            direction,
+                    GTlsOpensslIOFunc       perform_func,
+                    gpointer                perform_data,
+                    gint64                  timeout,
+                    GCancellable           *cancellable,
+                    int                    *out_ret,
+                    GError                **error,
+                    const char             *err_prefix)
+{
+  GTlsConnectionBaseStatus status;
+  GTlsConnectionBase *tls;
+  GTlsConnectionOpensslPrivate *priv;
+  SSL *ssl;
+  gint64 deadline;
+  int ret;
+
+  tls = G_TLS_CONNECTION_BASE (openssl);
+  priv = g_tls_connection_openssl_get_instance_private (openssl);
+  ssl = g_tls_connection_openssl_get_ssl (openssl);
+
+  if (timeout >= 0)
+    deadline = g_get_monotonic_time () + timeout;
+  else
+    deadline = -1;
+
+  while (TRUE)
+    {
+      GIOCondition io_needed;
+      char error_str[256];
+      struct timeval tv;
+      gint64 io_timeout;
+
+      g_tls_connection_base_push_io (tls, direction, 0, cancellable);
+
+      if (g_tls_connection_base_is_dtls (tls))
+        DTLSv1_handle_timeout (ssl);
+
+      ret = perform_func (ssl, perform_data);
+
+      switch (SSL_get_error (ssl, ret))
+        {
+          case SSL_ERROR_WANT_READ:
+            io_needed = G_IO_IN;
+            break;
+          case SSL_ERROR_WANT_WRITE:
+            io_needed = G_IO_OUT;
+            break;
+          default:
+            io_needed = 0;
+            break;
+        }
+
+      ERR_error_string_n (SSL_get_error (ssl, ret), error_str,
+                          sizeof (error_str));
+      status = end_openssl_io (openssl, direction, ret, TRUE, error, err_prefix,
+                               error_str);
+
+      if (status != G_TLS_CONNECTION_BASE_TRY_AGAIN)
+        break;
+
+      if (g_tls_connection_base_is_dtls (tls) && DTLSv1_get_timeout (ssl, &tv))
+        io_timeout = (tv.tv_sec * G_USEC_PER_SEC) + tv.tv_usec;
+      else
+        io_timeout = -1;
+
+      if (deadline != -1)
+        {
+          gint64 remaining = MAX (deadline - g_get_monotonic_time (), 0);
+
+          if (io_timeout != -1)
+            io_timeout = MIN (io_timeout, remaining);
+          else
+            io_timeout = remaining;
+        }
+
+      if (io_timeout == 0)
+        break;
+
+      g_tls_bio_wait_available (priv->bio, io_needed, io_timeout, cancellable);
+    }
+
+  if (status == G_TLS_CONNECTION_BASE_TRY_AGAIN)
+    {
+      if (timeout == 0)
+        {
+          status = G_TLS_CONNECTION_BASE_WOULD_BLOCK;
+          g_clear_error (error);
+          g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK,
+                               "Operation would block");
+        }
+      else if (timeout > 0)
+        {
+          status = G_TLS_CONNECTION_BASE_TIMED_OUT;
+          g_clear_error (error);
+          g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_TIMED_OUT,
+                               _("Socket I/O timed out"));
+        }
+    }
+
+  if (out_ret)
+    *out_ret = ret;
+
+  return status;
+}
+
 static int
 _openssl_alpn_select_cb (SSL                  *ssl,
                          const unsigned char **out,
@@ -359,13 +482,75 @@ g_tls_connection_openssl_prepare_handshake (GTlsConnectionBase  *tls,
     }
 }
 
+static GTlsCertificateFlags
+g_tls_connection_openssl_verify_chain (GTlsConnectionBase       *tls,
+                                       GTlsCertificate          *chain,
+                                       const gchar              *purpose,
+                                       GSocketConnectable       *identity,
+                                       GTlsInteraction          *interaction,
+                                       GTlsDatabaseVerifyFlags   flags,
+                                       GCancellable             *cancellable,
+                                       GError                  **error)
+{
+  GTlsDatabase *database;
+  GTlsCertificateFlags errors = 0;
+  gboolean is_client = G_IS_TLS_CLIENT_CONNECTION (tls);
+
+  database = g_tls_connection_get_database (G_TLS_CONNECTION (tls));
+  if (database)
+    {
+      errors |= g_tls_database_verify_chain (database,
+                                             chain,
+                                             is_client ? G_TLS_DATABASE_PURPOSE_AUTHENTICATE_SERVER : G_TLS_DATABASE_PURPOSE_AUTHENTICATE_CLIENT,
+                                             identity,
+                                             g_tls_connection_get_interaction (G_TLS_CONNECTION (tls)),
+                                             G_TLS_DATABASE_VERIFY_NONE,
+                                             NULL,
+                                             error);
+    }
+  else
+    {
+      errors |= G_TLS_CERTIFICATE_UNKNOWN_CA;
+      errors |= g_tls_certificate_verify (chain, identity, NULL);
+    }
+
+  return errors;
+}
+
+static GTlsProtocolVersion
+glib_protocol_version_from_openssl (int protocol_version)
+{
+  switch (protocol_version)
+    {
+    case SSL3_VERSION:
+      return G_TLS_PROTOCOL_VERSION_SSL_3_0;
+    case TLS1_VERSION:
+      return G_TLS_PROTOCOL_VERSION_TLS_1_0;
+    case TLS1_1_VERSION:
+      return G_TLS_PROTOCOL_VERSION_TLS_1_1;
+    case TLS1_2_VERSION:
+      return G_TLS_PROTOCOL_VERSION_TLS_1_2;
+    case TLS1_3_VERSION:
+      return G_TLS_PROTOCOL_VERSION_TLS_1_3;
+    case DTLS1_VERSION:
+      return G_TLS_PROTOCOL_VERSION_DTLS_1_0;
+    case DTLS1_2_VERSION:
+      return G_TLS_PROTOCOL_VERSION_DTLS_1_2;
+    default:
+      return G_TLS_PROTOCOL_VERSION_UNKNOWN;
+    }
+}
+
 static void
-g_tls_connection_openssl_complete_handshake (GTlsConnectionBase  *tls,
-                                             gboolean             handshake_succeeded,
-                                             gchar              **negotiated_protocol,
-                                             GError             **error)
+g_tls_connection_openssl_complete_handshake (GTlsConnectionBase   *tls,
+                                             gboolean              handshake_succeeded,
+                                             gchar               **negotiated_protocol,
+                                             GTlsProtocolVersion  *protocol_version,
+                                             gchar               **ciphersuite_name,
+                                             GError              **error)
 {
   SSL *ssl;
+  SSL_SESSION *session;
   unsigned int len = 0;
   const unsigned char *data = NULL;
 
@@ -373,6 +558,7 @@ g_tls_connection_openssl_complete_handshake (GTlsConnectionBase  *tls,
     return;
 
   ssl = g_tls_connection_openssl_get_ssl (G_TLS_CONNECTION_OPENSSL (tls));
+  session = SSL_get_session (ssl);
 
   SSL_get0_alpn_selected (ssl, &data, &len);
 
@@ -383,44 +569,18 @@ g_tls_connection_openssl_complete_handshake (GTlsConnectionBase  *tls,
       g_assert (!*negotiated_protocol);
       *negotiated_protocol = g_strndup ((gchar *)data, len);
     }
+
+  *protocol_version = glib_protocol_version_from_openssl (SSL_SESSION_get_protocol_version (session));
+  *ciphersuite_name = g_strdup (SSL_get_cipher_name (ssl));
 }
-#endif
 
-#define BEGIN_OPENSSL_IO(openssl, direction, timeout, cancellable)          \
-  do {                                                                      \
-    char error_str[256];                                                    \
-    g_tls_connection_base_push_io (G_TLS_CONNECTION_BASE (openssl),         \
-                                   direction, timeout, cancellable);
-
-#define END_OPENSSL_IO(openssl, direction, ret, timeout, status, errmsg, err) \
-    ERR_error_string_n (SSL_get_error (ssl, ret), error_str, sizeof(error_str)); \
-    status = end_openssl_io (openssl, direction, ret, timeout == -1, err, errmsg, error_str); \
-  } while (status == G_TLS_CONNECTION_BASE_TRY_AGAIN);
-
-static GTlsConnectionBaseStatus
-g_tls_connection_openssl_handshake_thread_request_rehandshake (GTlsConnectionBase  *tls,
-                                                               gint64               timeout,
-                                                               GCancellable        *cancellable,
-                                                               GError             **error)
+static int
+perform_rehandshake (SSL      *ssl,
+                     gpointer  user_data)
 {
-  GTlsConnectionOpenssl *openssl;
-  GTlsConnectionBaseStatus status;
-  SSL *ssl;
+  GTlsConnectionBase *tls = user_data;
   int ret = 1; /* always look on the bright side of life */
 
-  /* On a client-side connection, SSL_renegotiate() itself will start
-   * a rehandshake, so we only need to do something special here for
-   * server-side connections.
-   */
-  if (!G_IS_TLS_SERVER_CONNECTION (tls))
-    return G_TLS_CONNECTION_BASE_OK;
-
-  openssl = G_TLS_CONNECTION_OPENSSL (tls);
-
-  ssl = g_tls_connection_openssl_get_ssl (openssl);
-
-  BEGIN_OPENSSL_IO (openssl, G_IO_IN | G_IO_OUT, timeout, cancellable);
-
 #if OPENSSL_VERSION_NUMBER >= 0x10101000L
   if (SSL_version(ssl) >= TLS1_3_VERSION)
     ret = SSL_key_update (ssl, SSL_KEY_UPDATE_REQUESTED);
@@ -433,10 +593,25 @@ g_tls_connection_openssl_handshake_thread_request_rehandshake (GTlsConnectionBas
   ret = SSL_renegotiate (ssl);
 #endif
 
-  END_OPENSSL_IO (openssl, G_IO_IN | G_IO_OUT, ret, timeout, status,
-                  _("Error performing TLS handshake"), error);
+  return ret;
+}
 
-  return status;
+static GTlsConnectionBaseStatus
+g_tls_connection_openssl_handshake_thread_request_rehandshake (GTlsConnectionBase  *tls,
+                                                               gint64               timeout,
+                                                               GCancellable        *cancellable,
+                                                               GError             **error)
+{
+  /* On a client-side connection, SSL_renegotiate() itself will start
+   * a rehandshake, so we only need to do something special here for
+   * server-side connections.
+   */
+  if (!G_IS_TLS_SERVER_CONNECTION (tls))
+    return G_TLS_CONNECTION_BASE_OK;
+
+  return perform_openssl_io (G_TLS_CONNECTION_OPENSSL (tls), G_IO_IN | G_IO_OUT,
+                             perform_rehandshake, tls, timeout, cancellable,
+                             NULL, error, _("Error performing TLS handshake"));
 }
 
 static GTlsCertificate *
@@ -638,21 +813,18 @@ g_tls_connection_openssl_handshake_thread_handshake (GTlsConnectionBase  *tls,
                                                      GCancellable        *cancellable,
                                                      GError             **error)
 {
-  GTlsConnectionOpenssl *openssl = G_TLS_CONNECTION_OPENSSL (tls);
   GTlsConnectionBaseStatus status;
-  SSL *ssl;
   int ret;
 
-  ssl = g_tls_connection_openssl_get_ssl (openssl);
-
-  BEGIN_OPENSSL_IO (openssl, G_IO_IN | G_IO_OUT, timeout, cancellable);
-  ret = SSL_do_handshake (ssl);
-  END_OPENSSL_IO (openssl, G_IO_IN | G_IO_OUT, ret, timeout, status,
-                  _("Error performing TLS handshake"), error);
+  status = perform_openssl_io (G_TLS_CONNECTION_OPENSSL (tls),
+                               G_IO_IN | G_IO_OUT,
+                               (GTlsOpensslIOFunc) SSL_do_handshake,
+                               NULL, timeout, cancellable, &ret, error,
+                               _("Error reading data from TLS socket"));
 
   if (ret > 0)
     {
-      if (!g_tls_connection_base_handshake_thread_verify_certificate (G_TLS_CONNECTION_BASE (openssl)))
+      if (!g_tls_connection_base_handshake_thread_verify_certificate (tls))
         {
           g_set_error_literal (error, G_TLS_ERROR, G_TLS_ERROR_BAD_CERTIFICATE,
                                _("Unacceptable TLS certificate"));
@@ -678,14 +850,10 @@ g_tls_connection_openssl_push_io (GTlsConnectionBase *tls,
   G_TLS_CONNECTION_BASE_CLASS (g_tls_connection_openssl_parent_class)->push_io (tls, direction,
                                                                                 timeout, cancellable);
 
-  /* FIXME: need to support timeout > 0
-   * This will require changes in GTlsBio */
-
   if (direction & G_IO_IN)
     {
       error = g_tls_connection_base_get_read_error (tls);
       g_tls_bio_set_read_cancellable (priv->bio, cancellable);
-      g_tls_bio_set_read_blocking (priv->bio, timeout == -1);
       g_clear_error (error);
       g_tls_bio_set_read_error (priv->bio, error);
     }
@@ -694,7 +862,6 @@ g_tls_connection_openssl_push_io (GTlsConnectionBase *tls,
     {
       error = g_tls_connection_base_get_write_error (tls);
       g_tls_bio_set_write_cancellable (priv->bio, cancellable);
-      g_tls_bio_set_write_blocking (priv->bio, timeout == -1);
       g_clear_error (error);
       g_tls_bio_set_write_error (priv->bio, error);
     }
@@ -725,6 +892,15 @@ g_tls_connection_openssl_pop_io (GTlsConnectionBase  *tls,
                                                                                       success, error);
 }
 
+static int
+perform_read (SSL      *ssl,
+              gpointer  user_data)
+{
+  ReadRequest *req = user_data;
+
+  return SSL_read (ssl, req->buffer, req->count);
+}
+
 static GTlsConnectionBaseStatus
 g_tls_connection_openssl_read (GTlsConnectionBase    *tls,
                                void                  *buffer,
@@ -734,44 +910,76 @@ g_tls_connection_openssl_read (GTlsConnectionBase    *tls,
                                GCancellable          *cancellable,
                                GError               **error)
 {
-  GTlsConnectionOpenssl *openssl = G_TLS_CONNECTION_OPENSSL (tls);
-  GTlsConnectionOpensslPrivate *priv;
   GTlsConnectionBaseStatus status;
-  SSL *ssl;
-  gssize ret;
+  ReadRequest req = { buffer, count };
+  int ret;
 
-  priv = g_tls_connection_openssl_get_instance_private (openssl);
-
-  ssl = g_tls_connection_openssl_get_ssl (openssl);
-
-  /* FIXME: revert back to use BEGIN/END_OPENSSL_IO once we move all the ssl
-   * operations into a worker thread
-   */
-  while (TRUE)
-    {
-      char error_str[256];
-
-      /* We want to always be non blocking here to avoid deadlocks */
-      g_tls_connection_base_push_io (G_TLS_CONNECTION_BASE (openssl),
-                                     G_IO_IN, 0, cancellable);
-
-      ret = SSL_read (ssl, buffer, count);
-
-      ERR_error_string_n (SSL_get_error (ssl, ret), error_str, sizeof (error_str));
-      status = end_openssl_io (openssl, G_IO_IN, ret, timeout == -1, error,
-                               _("Error reading data from TLS socket"), error_str);
-
-      if (status != G_TLS_CONNECTION_BASE_TRY_AGAIN)
-        break;
-
-      /* Wait for the socket to be available again to avoid an infinite loop */
-      g_tls_bio_wait_available (priv->bio, G_IO_IN, cancellable);
-    }
+  status = perform_openssl_io (G_TLS_CONNECTION_OPENSSL (tls), G_IO_IN,
+                               perform_read, &req, timeout, cancellable, &ret,
+                               error, _("Error reading data from TLS socket"));
 
   *nread = MAX (ret, 0);
   return status;
 }
 
+static GTlsConnectionBaseStatus
+g_tls_connection_openssl_read_message (GTlsConnectionBase  *tls,
+                                       GInputVector        *vectors,
+                                       guint                num_vectors,
+                                       gint64               timeout,
+                                       gssize              *nread,
+                                       GCancellable        *cancellable,
+                                       GError             **error)
+{
+  GTlsConnectionOpenssl *openssl = G_TLS_CONNECTION_OPENSSL (tls);
+  GTlsConnectionOpensslPrivate *priv;
+  GTlsConnectionBaseStatus status;
+  gssize bytes_read;
+  gsize bytes_copied, bytes_remaining;
+  guint i;
+
+  *nread = 0;
+
+  priv = g_tls_connection_openssl_get_instance_private (openssl);
+
+  if (!priv->dtls_rx)
+    priv->dtls_rx = g_malloc (DTLS_MESSAGE_MAX_SIZE);
+
+  status = g_tls_connection_openssl_read (tls, priv->dtls_rx,
+                                          DTLS_MESSAGE_MAX_SIZE, timeout,
+                                          &bytes_read, cancellable, error);
+  if (status != G_TLS_CONNECTION_BASE_OK)
+    return status;
+
+  bytes_copied = 0;
+  bytes_remaining = bytes_read;
+  for (i = 0; i < num_vectors && bytes_remaining > 0; i++)
+    {
+      GInputVector *vector = &vectors[i];
+      gsize n;
+
+      n = MIN (bytes_remaining, vector->size);
+
+      memcpy (vector->buffer, priv->dtls_rx + bytes_copied, n);
+
+      bytes_copied += n;
+      bytes_remaining -= n;
+    }
+
+  *nread = bytes_copied;
+
+  return status;
+}
+
+static int
+perform_write (SSL      *ssl,
+               gpointer  user_data)
+{
+  WriteRequest *req = user_data;
+
+  return SSL_write (ssl, req->buffer, req->count);
+}
+
 static GTlsConnectionBaseStatus
 g_tls_connection_openssl_write (GTlsConnectionBase    *tls,
                                 const void            *buffer,
@@ -781,41 +989,56 @@ g_tls_connection_openssl_write (GTlsConnectionBase    *tls,
                                 GCancellable          *cancellable,
                                 GError               **error)
 {
-  GTlsConnectionOpenssl *openssl = G_TLS_CONNECTION_OPENSSL (tls);
-  GTlsConnectionOpensslPrivate *priv;
   GTlsConnectionBaseStatus status;
-  SSL *ssl;
-  gssize ret;
+  WriteRequest req = { buffer, count };
+  int ret;
 
-  priv = g_tls_connection_openssl_get_instance_private (openssl);
-
-  ssl = g_tls_connection_openssl_get_ssl (openssl);
-
-  while (TRUE)
-    {
-      char error_str[256];
-
-      /* We want to always be non blocking here to avoid deadlocks */
-      g_tls_connection_base_push_io (G_TLS_CONNECTION_BASE (openssl),
-                                     G_IO_OUT, 0, cancellable);
-
-      ret = SSL_write (ssl, buffer, count);
-
-      ERR_error_string_n (SSL_get_error (ssl, ret), error_str, sizeof (error_str));
-      status = end_openssl_io (openssl, G_IO_OUT, ret, timeout == -1, error,
-                               _("Error writing data to TLS socket"), error_str);
-
-      if (status != G_TLS_CONNECTION_BASE_TRY_AGAIN)
-        break;
-
-      /* Wait for the socket to be available again to avoid an infinite loop */
-      g_tls_bio_wait_available (priv->bio, G_IO_OUT, cancellable);
-    }
+  status = perform_openssl_io (G_TLS_CONNECTION_OPENSSL (tls), G_IO_OUT,
+                               perform_write, &req, timeout, cancellable, &ret,
+                               error, _("Error writing data to TLS socket"));
 
   *nwrote = MAX (ret, 0);
   return status;
 }
 
+static GTlsConnectionBaseStatus
+g_tls_connection_openssl_write_message (GTlsConnectionBase  *tls,
+                                        GOutputVector       *vectors,
+                                        guint                num_vectors,
+                                        gint64               timeout,
+                                        gssize              *nwrote,
+                                        GCancellable        *cancellable,
+                                        GError             **error)
+{
+  GTlsConnectionOpenssl *openssl = G_TLS_CONNECTION_OPENSSL (tls);
+  GTlsConnectionOpensslPrivate *priv;
+  gsize bytes_copied, bytes_available;
+  guint i;
+
+  priv = g_tls_connection_openssl_get_instance_private (openssl);
+
+  if (!priv->dtls_tx)
+    priv->dtls_tx = g_malloc (DTLS_MESSAGE_MAX_SIZE);
+
+  bytes_copied = 0;
+  bytes_available = DTLS_MESSAGE_MAX_SIZE;
+  for (i = 0; i < num_vectors && bytes_available > 0; i++)
+    {
+      GOutputVector *vector = &vectors[i];
+      gsize n;
+
+      n = MIN (vector->size, bytes_available);
+
+      memcpy (priv->dtls_tx + bytes_copied, vector->buffer, n);
+
+      bytes_copied += n;
+      bytes_available -= n;
+    }
+
+  return g_tls_connection_openssl_write (tls, priv->dtls_tx, bytes_copied,
+                                         timeout, nwrote, cancellable, error);
+}
+
 static GTlsConnectionBaseStatus
 g_tls_connection_openssl_close (GTlsConnectionBase  *tls,
                                 gint64               timeout,
@@ -824,25 +1047,16 @@ g_tls_connection_openssl_close (GTlsConnectionBase  *tls,
 {
   GTlsConnectionOpenssl *openssl = G_TLS_CONNECTION_OPENSSL (tls);
   GTlsConnectionOpensslPrivate *priv;
-  GTlsConnectionBaseStatus status;
-  SSL *ssl;
-  int ret;
 
-  ssl = g_tls_connection_openssl_get_ssl (openssl);
   priv = g_tls_connection_openssl_get_instance_private (openssl);
 
   priv->shutting_down = TRUE;
 
-  BEGIN_OPENSSL_IO (openssl, G_IO_IN | G_IO_OUT, timeout, cancellable);
-  ret = SSL_shutdown (ssl);
-  /* Note it is documented that getting 0 is correct when shutting down since
-   * it means it will close the write direction
-   */
-  ret = ret == 0 ? 1 : ret;
-  END_OPENSSL_IO (openssl, G_IO_IN | G_IO_OUT, ret, timeout, status,
-                  _("Error performing TLS close"), error);
-
-  return status;
+  return perform_openssl_io (G_TLS_CONNECTION_OPENSSL (tls),
+                             G_IO_IN | G_IO_OUT,
+                             (GTlsOpensslIOFunc) SSL_shutdown,
+                             NULL, timeout, cancellable, NULL, error,
+                             _("Error performing TLS close"));
 }
 
 static void
@@ -853,10 +1067,9 @@ g_tls_connection_openssl_class_init (GTlsConnectionOpensslClass *klass)
 
   object_class->finalize                                 = g_tls_connection_openssl_finalize;
 
-#if OPENSSL_VERSION_NUMBER >= 0x10002000L || defined (LIBRESSL_VERSION_NUMBER)
   base_class->prepare_handshake                          = g_tls_connection_openssl_prepare_handshake;
+  base_class->verify_chain                               = g_tls_connection_openssl_verify_chain;
   base_class->complete_handshake                         = g_tls_connection_openssl_complete_handshake;
-#endif
   base_class->handshake_thread_safe_renegotiation_status = g_tls_connection_openssl_handshake_thread_safe_renegotiation_status;
   base_class->handshake_thread_request_rehandshake       = g_tls_connection_openssl_handshake_thread_request_rehandshake;
   base_class->handshake_thread_handshake                 = g_tls_connection_openssl_handshake_thread_handshake;
@@ -865,7 +1078,9 @@ g_tls_connection_openssl_class_init (GTlsConnectionOpensslClass *klass)
   base_class->push_io                                    = g_tls_connection_openssl_push_io;
   base_class->pop_io                                     = g_tls_connection_openssl_pop_io;
   base_class->read_fn                                    = g_tls_connection_openssl_read;
+  base_class->read_message_fn                            = g_tls_connection_openssl_read_message;
   base_class->write_fn                                   = g_tls_connection_openssl_write;
+  base_class->write_message_fn                           = g_tls_connection_openssl_write_message;
   base_class->close_fn                                   = g_tls_connection_openssl_close;
 }
 
@@ -880,12 +1095,16 @@ g_tls_connection_openssl_initable_init (GInitable     *initable,
   GTlsConnectionOpensslPrivate *priv;
   GTlsConnectionBase *tls = G_TLS_CONNECTION_BASE (initable);
   GIOStream *base_io_stream;
+  GDatagramBased *base_socket;
   SSL *ssl;
 
   g_object_get (tls,
                 "base-io-stream", &base_io_stream,
+                "base-socket", &base_socket,
                 NULL);
-  g_return_val_if_fail (base_io_stream, FALSE);
+
+  /* Ensure we are in TLS mode or DTLS mode. */
+  g_return_val_if_fail (!!base_io_stream != !!base_socket, FALSE);
 
   priv = g_tls_connection_openssl_get_instance_private (openssl);
 
@@ -897,11 +1116,15 @@ g_tls_connection_openssl_initable_init (GInitable     *initable,
   }
   SSL_set_ex_data (ssl, data_index, openssl);
 
-  priv->bio = g_tls_bio_new (base_io_stream);
+  if (base_io_stream)
+    priv->bio = g_tls_bio_new_from_iostream (base_io_stream);
+  else
+    priv->bio = g_tls_bio_new_from_datagram_based (base_socket);
 
   SSL_set_bio (ssl, priv->bio, priv->bio);
 
-  g_object_unref (base_io_stream);
+  g_clear_object (&base_io_stream);
+  g_clear_object (&base_socket);
 
   return TRUE;
 }
